import java.awt.Color;
import java.awt.Point;
import java.lang.IllegalArgumentException;

/**
 * Eine Bodenkachel im Labyrinth.
 * 
 * @author Bernd Gramlich
 * @version 2013-06-21
 */
public class Bodenkachel extends Kachel
{
    /*
     * Die Koordinaten des Kachelmittelpunkts. Diesen
     * brauchen wir, um Marken auf den Boden zu zeichnen.
     */
    protected int xm;
    protected int ym;
    
    /**
     * Konstruktor.
     * Erzeuge eine Bodenkachel an der angegebenen Position mit
     * der angegebenen Faerbung. Die Ausgabeleinwand hat dabei
     * die angegebene Hoehe und Breite und die angegebenen
     * Fluchtpunktkoordinaten.
     * 
     * Eine Bodenkachel hat immer geradzahlige Koordinaten. Die
     * y-Koordinate ist nicht negativ. Wenn die Argumente x und y
     * diese Bedingung nicht erfuellen, melde einen Fehler.
     */
    public Bodenkachel (int x, int y, Color dieseFaerbung, int breite, int hoehe, int xf, int yf, double mu, double weite)
        throws IllegalArgumentException
    {
        super (x, y, dieseFaerbung, breite, hoehe, xf, yf, mu, weite);
        
        /*
         * Wenn eine der beiden Koordinaten ungerade ist, melde einen Fehler.
         */
        if (x % 2 == 1 || y % 2 == 1)
        {
            throw new IllegalArgumentException ("Bodenkacheln haben geradzahlige x- und y-Koordinaten!");
        }
        
        /*
         * Wenn die y-Koordinate negativ ist, melde einen Fehler.
         */
        if (y < 0)
        {
            throw new IllegalArgumentException ("Bodenkacheln haben nichtnegative y-Koordinate!");
        }
        
        abstand = Math.abs (x) + y + 1;
        
        euklidischerAbstand = Math.sqrt (Math.pow (x, 2) + Math.pow (y, 2) + 1);
        
        dP = Math.sqrt (Math.pow (x, 2) + Math.pow (y - 1, 2) + 1);
        dQ = Math.sqrt (Math.pow (x, 2) + Math.pow (y + 1, 2) + 1);
        zyklisch = false;
        
        aendereLeinwanddaten (dieseFaerbung, breite, hoehe, xf, yf, mu, weite);
    }
    
    /**
     * Aendere die Leinwanddaten.
     * Die Bodenkachel muss entsprechend den neuen Leinwanddaten geformt
     * und gefaerbt werden.
     */
    public void aendereLeinwanddaten (Color dieseFaerbung, int breite, int hoehe, int xf, int yf, double mu, double weite)
    {
        double lambda = mu * (2 - mu);
        
        bild.loescheAlleEcken ();
        
        int dx = positionX / 2;
        int dy = positionY / 2;
        
        double faktoru = 1 - (1 - mu) * Math.pow (1 - lambda, dy - 1);
        double faktorm = 1 - Math.pow (1 - lambda, dy);
        double faktoro = 1 - (1 - mu) * Math.pow (1 - lambda, dy);
        
        if (0 == dy)
        {
            faktoru = 0;
        }
            
        int xlu = (int) Math.round (faktoru * xf);
        int xlm = (int) Math.round (faktorm * xf);
        int xlo = (int) Math.round (faktoro * xf);
        int xru = (int) (breite - Math.round (faktoru * (breite - xf)));
        int xrm = (int) (breite - Math.round (faktorm * (breite - xf)));
        int xro = (int) (breite - Math.round (faktoro * (breite - xf)));
        int yu  = (int) (hoehe - Math.round (faktoru * (hoehe - yf)));
            ym  = (int) (hoehe - Math.round (faktorm * (hoehe - yf)));
        int yo  = (int) (hoehe - Math.round (faktoro * (hoehe - yf)));
            
        bild.ergaenzeEcke (dx * (xru - xlu) + xlu, yu);
        bild.ergaenzeEcke (dx * (xru - xlu) + xru, yu);
        bild.ergaenzeEcke (dx * (xro - xlo) + xro, yo);
        bild.ergaenzeEcke (dx * (xro - xlo) + xlo, yo);
        
        xm = dx * (xrm - xlm) + (xrm + xlm)/2 /*(int) (breite/2 - Math.round (faktorm * (breite/2 - xf)))*/;
        
        P = new Point (dx * (xru - xlu) + (xru + xlu)/2, yu);
        Q = new Point (dx * (xro - xlo) + (xro + xlo)/2, yo);
                
        hellsichtweite = weite;
        faerbung = dieseFaerbung;
        bild.setzeFarbe (erzeugeBildfaerbung ());
    }
    
    /**
     * Pruefe, ob die Kachel eine Wandkachel ist.
     */
    public boolean istWandkachel ()
    {
        return false;
    }
}
