import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.GridLayout;
import java.io.File;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.JButton;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JSlider;
import javax.swing.SwingConstants;

/**
 * Ein Dialogfenster fuer das Labyrinthspiel.
 * Vorgesehen sind je ein Button fuer das Vorwaertsgehen, 
 * das Rueckwaertsgehen, das Drehen nach Links und nach Rechts und das Setzen
 * bzw. Loeschen von Marken, ein Button, um ein neues
 * Labyrinth zu laden, ein Button, um ein Labyrinth zu speichern, ein Helligkeitsregler und ein Label,
 * in dem (hauptsaechlich) die Anzahl der noch im Labyrinth
 * vorhandenen Schaetze angezeigt wird.
 * 
 * @author Bernd Gramlich
 * @author Felicia W.
 * @author Stefanie E.
 * @author Romy B.
 * @author Laura T.
 * @version 2013-05-17
 */
public class Dialogfenster implements PersonenBeobachter, ActionListener, KeyListener
{
    /*
     * Das Labyrinthmodell, das mich mit Labyrinthdaten
     * versorgt.
     */
    private LabyrinthModel model;
    
    /*
     * Die Labyrinthsteuerung, die mich verwaltet.
     */
    private LabyrinthController controller;
    
    /*
     * Die Komponenten des Dialogfensters.
     */
    private JFrame dialogfenster;
    private JPanel dialograum;
    private JPanel linksrechtsraum;
    private JButton vorwaertsknopf;
    private JButton linksknopf;
    private JButton rechtsknopf;
    private JButton markenknopf;
    private JLabel anzeige;
    private JButton ladenknopf;
    private JButton speichernknopf;
    private JButton rueckwaertsknopf;
    private Bildfenster bildfenster;
    
    private String naechstesBild;
    
    /*
     * Der Auswahldialog zum Laden eines Labyrinths.
     */
    private JFileChooser dateiauswahl;

    /**
     * Konstruktor.
     * Erzeuge ein Dialogfenster mit dem LabyrinthModel m
     * und dem LabyrinthController c.
     */
    public Dialogfenster (LabyrinthModel m, LabyrinthController c)
    {
        model      = m;
        controller = c;

        model.speichereAlsBeobachter (this);
        
        vorwaertsknopf  = new JButton ("^");
        rueckwaertsknopf = new JButton ("v");
        linksknopf      = new JButton ("<");
        rechtsknopf     = new JButton (">");
        markenknopf     = new JButton ("Marke");
        ladenknopf      = new JButton ("Labyrinth laden");
        speichernknopf  = new JButton ("Labyrinth speichern");
        vorwaertsknopf.addActionListener (this);
        rueckwaertsknopf.addActionListener (this);
        linksknopf.addActionListener (this);
        rechtsknopf.addActionListener (this);
        markenknopf.addActionListener (this);
        ladenknopf.addActionListener (this);
        speichernknopf.addActionListener (this);
        vorwaertsknopf.addKeyListener (this);
        rueckwaertsknopf.addKeyListener (this);
        linksknopf.addKeyListener (this);
        rechtsknopf.addKeyListener (this);
        markenknopf.addKeyListener (this);
        ladenknopf.addKeyListener (this);
        speichernknopf.addKeyListener (this);
        
        anzeige   = new JLabel ("Lade zuerst ein Labyrinth.");
        anzeige.setHorizontalAlignment (SwingConstants.CENTER);
        
        
        linksrechtsraum = new JPanel (new GridLayout (1, 2));
        linksrechtsraum.add (linksknopf);
        linksrechtsraum.add (rechtsknopf);
        
        dialograum      = new JPanel (new GridLayout (7, 1));
        dialograum.add (vorwaertsknopf);
        dialograum.add (rueckwaertsknopf);
        dialograum.add (linksrechtsraum);
        dialograum.add (markenknopf);
        dialograum.add (anzeige);
        dialograum.add (speichernknopf);
        dialograum.add (ladenknopf);
        
        dialogfenster   = new JFrame ("Labyrinthsteuerung");
        dialogfenster.setDefaultCloseOperation (JFrame.EXIT_ON_CLOSE);
        dialogfenster.getContentPane ().add (dialograum);
        dialogfenster.pack ();
        dialogfenster.setLocationRelativeTo (null);
        dialogfenster.setVisible (true);
        
        dateiauswahl = new JFileChooser ()
        {
            /*
             * Dieser Code entstammt dem Diskussionsforum stackoverflow.com gemaess
             * einem Vorschlag der Autoren rlbisbe und james.garriss .
             */
            @Override
            public void approveSelection()
            {
                // Sicherheitsfrage, ob es tatschlich berschrieben werden soll. 
            
                File f = getSelectedFile();
                if (f.exists () && getDialogType () == SAVE_DIALOG)
                {
                    int result = JOptionPane.showConfirmDialog (this, "Datei existiert. Soll die Datei berschrieben werden?", "Datei existiert.", JOptionPane.YES_NO_CANCEL_OPTION);
                    switch (result)
                    {
                        case JOptionPane.YES_OPTION:
                        super.approveSelection ();
                        return;
                        case JOptionPane.NO_OPTION:
                        return;
                        case JOptionPane.CLOSED_OPTION:
                        return;
                        case JOptionPane.CANCEL_OPTION:
                        cancelSelection ();
                        return;
                    }
                }
                super.approveSelection ();
            }
        };
        
        Leinwand.gibLeinwand ().gibFenster ().addKeyListener (this);
    }

    /**
     * Aktualisiere die Anzahl der Personen.
     */
    public void aktualisierePersonenzahl ()
    {
        int taeterzahl  = model.gibTaeterzahl ();
        int leichenzahl = model.gibLeichenzahl ();
        int zeugenzahl  = model.gibZeugenzahl ();
        int feindezahl  = model.gibFeindezahl ();
        
        String ausgabetext = "";
        
        if (taeterzahl < 0 && leichenzahl <0)
        {
            ausgabetext = " Spielende.";
            
            controller.ende ();
        }
        else
        {
            if (leichenzahl > 1)
            {
                ausgabetext = leichenzahl + " Entfhrte.";
            }
            else if (1 == leichenzahl)
            {
                ausgabetext = "1 Entfhrte.";
            }
            else if (0 == leichenzahl)
            {
                ausgabetext = "Keine Entfhrte.";
            }
            
            if  (1 == taeterzahl)
            {
                ausgabetext = ausgabetext + " 1 Tter.";
            }
            else if (0 == taeterzahl)
            {
                 ausgabetext = ausgabetext + " Kein Tter.";
            }
        }
    
        anzeige.setText (ausgabetext);
    }
    
    /**
     * Reagiere auf Benutzeraktionen.
     */
    public void actionPerformed (ActionEvent ereignis)
    {
        if (ereignis.getSource () == vorwaertsknopf)
        {
            controller.gehe ();
        }
        else if (ereignis.getSource () == rueckwaertsknopf)
        {
            controller.geheRueckwaerts ();
        }
        else if (ereignis.getSource () == linksknopf)
        {
            controller.dreheNachLinks ();
        }
        else if (ereignis.getSource () == rechtsknopf)
        {
            controller.dreheNachRechts ();
        }
        else if (ereignis.getSource () == markenknopf)
        {
            controller.markiere ();
        }
        else if (ereignis.getSource () == ladenknopf)
        {
            /*
             * Oeffne das Dateiauswahlfenster zum Einlesen des
             * Labyrinths. Falls der Benutzer eine Datei ausgewaehlt
             * hat ...
             */
            if (dateiauswahl.showOpenDialog (null) == JFileChooser.APPROVE_OPTION)
            { 
                /*
                 * ... lade das Labyrinth von dort.
                 */
                controller.ladeLabyrinth (dateiauswahl.getSelectedFile ());
            }
        }
        
        else if (ereignis.getSource () == speichernknopf)
        {
            /*
             * Oeffne das Dateiauswahlfenster zum Speichern des
             * Labyrinths. Falls der Benutzer eine Datei ausgewaehlt
             * hat ...
             */
            if (dateiauswahl.showSaveDialog (null) == JFileChooser.APPROVE_OPTION)
            { 
                /*
                 * ... speichern das Labyrinth von dort.
                 */
                controller.speichereLabyrinth (dateiauswahl.getSelectedFile ());
            }
        }
    }
    
    /**
     * Erlaube dem Benutzer, geradeaus zu gehen.
     */
    public void erlaubeGehen (boolean erlaubt)
    {
        vorwaertsknopf.setEnabled (erlaubt);
    }
    
    /**
     * Erlaube dem Benutzer, im Labyrinth rckwrts zu gehen.
     */
    public void erlaubeGeheRueckwaerts (boolean erlaubt)
    {
        rueckwaertsknopf.setEnabled (erlaubt);
    }
    
    /**
     * Erlaube dem Benutzer, sich im Labyrinth zu bewegen.
     */
    public void erlaubeBewegen (boolean erlaubt)
    {
        vorwaertsknopf.setEnabled (erlaubt);
        rueckwaertsknopf.setEnabled (erlaubt);
        linksknopf.setEnabled (erlaubt);
        rechtsknopf.setEnabled (erlaubt);
        markenknopf.setEnabled (erlaubt);
    }
    
    /**
     * Erlaube dem Benutzer, ein Labyrinth zu laden.
     */
    public void erlaubeLaden (boolean erlaubt)
    { 
        ladenknopf.setEnabled (erlaubt);
    }
    
    /**
     * Erlaube dem Benutzer, ein Labyrinth zu speichern.
     */
    public void erlaubeSpeichern (boolean erlaubt)
    { 
        speichernknopf.setEnabled (erlaubt);
    }
    
    /**
     * Reagiere auf Benutzeraktionen der Tastatursteuerung. Wenn eine relevante Taste gedrckt
     * und wieder losgelassen wurde, lse den zugehrigen Knopf aus
     */
    public void keyPressed (KeyEvent taste)
    {
        switch (taste.getKeyCode ())
        {
            case KeyEvent.VK_UP:    vorwaertsknopf.doClick(); break;
            case KeyEvent.VK_DOWN:  rueckwaertsknopf.doClick(); break;
            case KeyEvent.VK_LEFT:  linksknopf.doClick(); break;
            case KeyEvent.VK_RIGHT: rechtsknopf.doClick(); break;
        }
    }
    

    /**
     * Reagiere auf Benutzeraktionen der Tastatursteuerung, Wenn eine relevante Taste gedrckt
     * und wieder losgelassen wurde, lse den zugehrigen Knopf aus.
     */
    public void keyTyped (KeyEvent taste)
    {
        switch (taste.getKeyChar ())
        {
            case 'm': case 'M': markenknopf.doClick(); break;
            case 'l': case 'L': ladenknopf.doClick(); break;
            case 's': case 'S': speichernknopf.doClick(); break;
        }
    }
    
    
     /**
     * Reagiere auf Benutzeraktionen der Tastatursteuerung.
     */
    public void keyReleased (KeyEvent taste)
    {     
    }
   
    public void zeigeNaechstesBild ()
    {
        bildfenster = new Bildfenster (dialogfenster, "  ", model.zeigeNaechstesBild (), model.gibNaechstenInfotext () );
    }
}
