import java.io.File;
import java.util.ArrayList;

/**
 * Ein Labyrinth mit dicken Waenden.
 * Das Labyrinth wird als zweidimensionales, rechteckiges
 * Feld von Textzeichen gespeichert.
 * 
 * @author Bernd Gramlich
 * @author Felicia W.
 * @author Nicole W.
 * @author Stefanie E.
 * @author Romy B.
 * @author Inken B.
 * @author Laura T.
 * @version 2013-06-21
 */

public class DickwandigesLabyrinth implements LabyrinthModel
{
    /*
     * Die Codezeichen, die wir im Labyrinth erwarten.
     */
    private final char WAND         = '#';
    private final char LEERRAUM     = '.';
    private final char EINGANG      = '@';
    private final char MARKENORD    = '^';
    private final char MARKEWEST    = '<';
    private final char MARKESUED    = 'v';
    private final char MARKEOST     = '>';
    private final char POSITIONNORD = 'N';
    private final char POSITIONSUED = 'S';
    private final char POSITIONWEST = 'W';
    private final char POSITIONOST  = 'O';
    private final char FEIND        = 'F';
    private final char [] ZEUGE     = {'1', '2', '3', '4', '5', '6', '7', '8', '9'};
    private final char LEICHE       = 'L';
    private final char TAETER       = 'T';

    /*
     * Das Minimallabyrinth, das nur aus einem Eingang besteht.
     */
    private final char[][] minimalLabyrinth = {{EINGANG}};

    /*
     * Das dickwandige Labyrinth.
     */
    private Labyrinth labyrinth;

    /*
     * Die Breite und Hoehe des dickwandigen
     * Labyrinths.
     */
    private int breite;
    private int hoehe;

    /*
     * Die duennwandige, orientierte Ansicht des
     * Labyrinths, die wir an die Beobachter versenden.
     */
    private char[][] ansicht;

    /*
     * Die Seitenlaenge des quadratischen
     * Labyrinthausschnitts.
     */
    private int seitenlaenge;

    /*
     * Unsere Position im dickwandigen Labyrinth.
     */
    private int positionX;
    private int positionY;

    /*
     * Unsere Blickrichtung.
     */
    private Himmelsrichtung blickrichtung;

    /*
     * Die Anzahl der unentdeckten Tter.
     */
    private int taeter;

    /*
     * Die Anzahl der unentdeckten Leichen.
     */
    private int leichen;

    /*
     * Die Anzahl der unentdeckten Feinde.
     */
    private int feinde;

    /*
     * Die Anzahl der unentdeckten Zeugen.
     */
    private int zeugen;
    
    /*
     * Das naechste Bild, das wir in einem Bildfenster anzeigen.
     */
    private String naechstesBild;
    
    /*
     * Der Infotext, der im naechsten Bildfenster steht.
     */
    private String infotext;

    /*
     * Die Listen unserer Beobachter.
     */
    private ArrayList<PositionsBeobachter> positionsBeobachter;
    private ArrayList<PersonenBeobachter>  personenBeobachter;

    /**
     * Konstruktor.
     * Erzeuge ein Minimallabyrinth und siehe dafuer eine
     * Ansicht mit der genannten Seitenlaenge vor.
     */
    public DickwandigesLabyrinth (int dieseSeitenlaenge)
    {
        breite = 1;
        hoehe  = 1;

        labyrinth = new Labyrinth ();

        positionX = 0;
        positionY = 0;
        blickrichtung = Himmelsrichtung.NORD;

        taeter  = 0;
        feinde  = 0;
        leichen = 0;
        zeugen  = 0;

        seitenlaenge = dieseSeitenlaenge;

        ansicht = new char [4*seitenlaenge+1][4*seitenlaenge+1];

        positionsBeobachter = new ArrayList<PositionsBeobachter> ();
        personenBeobachter  = new ArrayList<PersonenBeobachter>  ();
    }

    /**
     * Gehe im Labyrinth einen Schritt geradeaus.
     */
    public void gehe ()
    {
        /*
         * Bewege Dich auf das naechste Feld geradeaus.
         */
        switch (blickrichtung)
        {
            case NORD: positionY++; break;
            case WEST: positionX--; break;
            case SUED: positionY--; break;
            case OST:  positionX++; break;
        }
        
        werteZielfeldAus ();
    }    
        
    /**
     * Gehe im Labyrinth einen Schritt rckwrts.
     */
    public void geheRueckwaerts ()
    {
        /*
         * Bewege Dich auf das naechste Feld rckwrts.
         */
        switch (blickrichtung)
        {
            case NORD: positionY--; break;
            case WEST: positionX++; break;
            case SUED: positionY++; break;
            case OST:  positionX--; break;
        }
        
        werteZielfeldAus ();
    }
    
    /**
     * berprfe, ob das Zielfeld beim Vorwrts- und Rckwrtsgehen ein besonderes Feld ist.
     */
    private void werteZielfeldAus ()
    {
        /*
         * Wenn das Feld ein Taeterfeld ist, nimm den
         * Tter (mache das Feld zum Leerraum).
         */
        if (TAETER == gibLabyrinthinhalt (positionX, positionY))
        {
            labyrinth.setzeFeld (positionX, positionY, LEERRAUM);
            taeter --;
            infotext = "Du hast den Tter gefunden";
            naechstesBild = ziehe (taeterbild, taeterzahl);
            benachrichtigePersonenBeobachter ();
        }

        /*
         * Wenn das Feld ein Leichenfeld ist, nimm die
         * Leiche (mache das Feld zum Leerraum).
         */
        if (LEICHE == gibLabyrinthinhalt (positionX, positionY))
        {
            labyrinth.setzeFeld (positionX, positionY, LEERRAUM);
            leichen --;
            infotext = "Du hast ein Opfer gefunden";
            naechstesBild = ziehe (opferbild, opferzahl);
            benachrichtigePersonenBeobachter ();
        }

        /*
         * Wenn das Feld ein Zeugenfeld ist, nimm den
         * Zeugen (mache das Feld zum Leerraum).
         * Auerdem soll die Information des Zeugens
         * in einem Extrafenster angezeigt werden.
         */
        for (int i = 0; i < 9; i++)
        {
            if (ZEUGE [i] == gibLabyrinthinhalt (positionX, positionY))
            {
                infotext = labyrinth.gibZeugenaussage (i+1);
                naechstesBild = ziehe (zeugenbild, zeugenzahl);
                benachrichtigePersonenBeobachter ();
            }
        } 

        /*
         * Wenn das Feld ein Feindfeld ist, nimm den
         * Feind (mache das Feld zum Leerraum).
         * Auerdem soll der Spieler zurck zum Eingang
         * geschickt werden.
         */
        if (FEIND == gibLabyrinthinhalt (positionX, positionY))
        {
            labyrinth.setzeFeld (positionX, positionY, LEERRAUM);
            feinde --;
            infotext = "Du hast einen Kumpanen gefunden";
            naechstesBild = ziehe (kumpanenbild, kumpanenzahl);
            benachrichtigePersonenBeobachter ();
            geheZumEingang ();
        }

        /*
         * Wenn alle Taeter und Leichen gefunden sind, loesche alle Zeugen und Feinde
         * aus dem Labyrinth.
         */
        if (0 == taeter && 0 == leichen)
        {
            for (int x = 0; x < hoehe; x++)
            {
                for (int y = 0; y < breite; y++)
                {
                    switch (labyrinth.gibFeld (x, y))
                    {
                        case FEIND:
                        labyrinth.setzeFeld (x, y, LEERRAUM);
                        break;
                    }

                    for (int k = 0; k < 9; k++)
                    {
                        if (labyrinth.gibFeld (x, y) == ZEUGE[k])
                        {
                            labyrinth.setzeFeld (x, y, LEERRAUM);
                        }
                    }
                }
            }
            feinde = 0;
            zeugen = 0;
        }

        /*
         * Erneuere die Labyrinthansicht.
         */
        erstelleAnsicht ();

        /*
         * Pruefe, ob Spielende erreicht.
         */
        pruefeSpielende ();
    }

    

    /**
     * Pruefe, ob ein Schritt geradeaus moeglich ist.
     * 
     * return True genau dann, wenn ein Schritt geradeaus moeglich ist.
     */
    public boolean istFrei ()
    {
        /*
         * Ich ermittle die Labyrinthkachel vor mir.
         */
        char kachelVorMir = WAND;

        switch (blickrichtung)
        {
            case NORD: kachelVorMir = gibLabyrinthinhalt (positionX    , positionY + 1); break;
            case WEST: kachelVorMir = gibLabyrinthinhalt (positionX - 1, positionY    ); break;
            case SUED: kachelVorMir = gibLabyrinthinhalt (positionX    , positionY - 1); break;
            case OST:  kachelVorMir = gibLabyrinthinhalt (positionX + 1, positionY    ); break;
        }

        /*
         * Melde, ob die Labyrinthkachel vor mir keine Wand ist.
         */
        return (kachelVorMir != WAND);
    }

    /**
     * Pruefe, ob ein Schritt rckwrts moeglich ist.
     * 
     * return True genau dann, wenn ein Schritt nach hinten moeglich ist.
     */
    public boolean istHintenFrei ()
    {
        /*
         * Ich ermittle die Labyrinthkachel hinter mir.
         */
        char kachelHinterMir = WAND;

        switch (blickrichtung)
        {
            case NORD: kachelHinterMir = gibLabyrinthinhalt (positionX    , positionY - 1); break;
            case WEST: kachelHinterMir = gibLabyrinthinhalt (positionX + 1, positionY    ); break;
            case SUED: kachelHinterMir = gibLabyrinthinhalt (positionX    , positionY + 1); break;
            case OST:  kachelHinterMir = gibLabyrinthinhalt (positionX - 1, positionY    ); break;
        }

        /*
         * Melde, ob die Labyrinthkachel hinter mir keine Wand ist.
         */
        return (kachelHinterMir != WAND);
    }

    /**
     * Drehe Dich im Labyrinth nach links.
     */
    public void dreheNachLinks ()
    {
        /*
         * Ich schwenke meinen Blick nach links.
         */
        switch (blickrichtung)
        {
            case NORD: blickrichtung = Himmelsrichtung.WEST; break;
            case WEST: blickrichtung = Himmelsrichtung.SUED; break;
            case SUED: blickrichtung = Himmelsrichtung.OST; break;
            case OST:  blickrichtung = Himmelsrichtung.NORD; break;
        }

        /*
         * Erneuere die Labyrinthansicht.
         */
        erstelleAnsicht ();

        /*
         * Pruefe, ob Spielende erreicht.
         */
        pruefeSpielende ();
    }

    /**
     * Drehe Dich im Labyrinth nach rechts.
     */
    public void dreheNachRechts ()
    {
        /*
         * Ich schwenke meinen Blick nach rechts.
         */
        switch (blickrichtung)
        {
            case NORD: blickrichtung = Himmelsrichtung.OST; break;
            case OST:  blickrichtung = Himmelsrichtung.SUED; break;
            case SUED: blickrichtung = Himmelsrichtung.WEST; break;
            case WEST: blickrichtung = Himmelsrichtung.NORD; break;
        }

        /*
         * Erneuere die Labyrinthansicht.
         */
        erstelleAnsicht ();

        /*
         * Pruefe, ob Spielende erreicht.
         */
        pruefeSpielende ();
    }

    /**
     * Setze an der augenblicklichen Position eine Marke.
     */
    public void setzeMarke ()
    {
        if (EINGANG != labyrinth.gibFeld (positionX, positionY))
        {
            /*
             * Falls ich nicht am Eingang bin, setze ich eine Marke, die
             * entgegen der Blickrichtung zeigt (also den Rueckweg weist).
             */
            switch (blickrichtung)
            {
                case NORD: labyrinth.setzeFeld (positionX, positionY, MARKESUED); break;
                case WEST: labyrinth.setzeFeld (positionX, positionY, MARKEOST);  break;
                case SUED: labyrinth.setzeFeld (positionX, positionY, MARKENORD); break;
                case OST:  labyrinth.setzeFeld (positionX, positionY, MARKEWEST); break;
            }

            /*
             * Erneuere die Labyrinthansicht.
             */
            erstelleAnsicht ();
        }
    }

    /**
     * Loesche die Marke an der augenblicklichen Position (sofern
     * dort eine ist).
     */
    public void loescheMarke ()
    {
        if (LEERRAUM != labyrinth.gibFeld (positionX, positionY))
        {
            /*
             * Wenn an der augenblicklichen Stelle kein Leerraum
             * ist, mache sie zum Leerraum.
             */
            labyrinth.setzeFeld (positionX, positionY, LEERRAUM);

            /*
             * Erneuere die Labyrinthansicht.
             */
            erstelleAnsicht ();
        }
    }

    /**
     * Pruefe, ob an der augenblicklichen Position eine Marke ist.
     * 
     * return True genau dann, wenn eine Marke vorhanden ist.
     */
    public boolean istMarke ()
    {
        boolean ausgabe = false;

        /*
         * Nur falls eine Marke vorhanden ist, die entgegen der Blickrichtung
         * zeigt, melden wir Erfolg.
         */
        switch (blickrichtung)
        {
            case NORD: ausgabe = (MARKESUED == labyrinth.gibFeld (positionX, positionY)); break;
            case WEST: ausgabe = (MARKEOST  == labyrinth.gibFeld (positionX, positionY)); break;
            case SUED: ausgabe = (MARKENORD == labyrinth.gibFeld (positionX, positionY)); break;
            case OST:  ausgabe = (MARKEWEST == labyrinth.gibFeld (positionX, positionY)); break;
        }

        return ausgabe;
    }

    /**
     * Speichere das Labyrinth in die Labyrinthdatei.
     * 
     * @param labyrinthDatei Die Datei, in die das Labyrinth gespeichert werden soll.
     */
    public boolean speichereLabyrinth (File labyrinthDatei)
    {
        return Datenschreiber.gibDatenschreiber ().schreibe (labyrinth, labyrinthDatei, positionX, positionY, blickrichtung);
    }

    /**
     * Lade das Labyrinth aus der Labyrinthdatei.
     * 
     * @param labyrinthDatei Die Datei, aus der das Labyrinth geladen werden soll.
     */
    public void ladeLabyrinth (File labyrinthDatei)
    {
        /*
         * Das Laden uebernimmt die Klasse Datenleser.
         */
        labyrinth = Datenleser.gibDatenleser ().lies (labyrinthDatei);

        /*
         * Wenn das Laden fehlschlaegt, geben wir keine
         * Fehlermeldung aus, sondern leben in einem
         * Minimallabyrinth.
         */
        if (null == labyrinth)
        {
            labyrinth = new Labyrinth ();
        }

        /*
         * Wir verlassen uns darauf, dass das geladene
         * Labyrinth rechteckig ist (oder vom Datenleser
         * rechteckig gemacht wurde).
         */
        hoehe  = labyrinth.gibHoehe ();
        breite = labyrinth.gibBreite ();

        /*
         * Ermittle den Eingang.
         */
        positionX = -1;
        positionY = -1;

        for (int x = 0; x < hoehe; x++)
        {
            for (int y = 0; y < breite; y++)
            {
                switch (labyrinth.gibFeld (x, y))
                {
                    case EINGANG:
                    if (positionX == -1 && positionY == -1)
                    {
                        positionX = x;
                        positionY = y;
                    }
                    break;
                }
            }
        }

        /*
         * Es folgt der Algorithmus nach Dijkstra.
         * 
         * Fr jedes Labyrinthkstchen ermitteln wir die Entfernung vom
         * Eingang und speichern, ob es schon besucht wurde.
         */
        double[][]  entfernung = new double[hoehe][breite];
        boolean[][] besucht    = new boolean[hoehe][breite];

        /*
         * Am Anfang habe alle Kstchen unendliche Entfernung vom Eingang und
         * sind noch nicht besucht.
         */
        for (int x = 0; x < hoehe; x++)
        {
            for (int y = 0; y < breite; y++)
            {
                entfernung[x][y] = Double.POSITIVE_INFINITY;
                besucht[x][y]    = false;
            }
        }

        /*
         * Der Eingang hat Entfernung 0 vom Eingang.
         */
        entfernung[positionX][positionY] = 0;

        /*
         * Die Koordinaten des aktuellen Kstchens.
         */
        int aktuellX = -1;
        int aktuellY = -1;

        /*
         * Die minimale Entfernung eines noch nicht besuchten Kstchens vom
         * Eingang.
         */
        double e;

        /*
         * Wiederhole ...
         */
        do
        {
            /*
             * Nimm als aktuelles Kaestchen dasjenige noch nicht besuchte
             * Kstchen, das am nchsten am Eingang ist.
             */
            e = Double.POSITIVE_INFINITY;

            for (int x = 0; x < hoehe; x++)
            {
                for (int y = 0; y < breite; y++)
                {
                    if (!besucht[x][y] && entfernung[x][y] < e)
                    {
                        e = entfernung[x][y];
                        aktuellX = x;
                        aktuellY = y;
                    }
                }
            }
            
            /*
             * Wenn die minimale Entfernung endlich ist ...
             */
            if (e < Double.POSITIVE_INFINITY)
            {
                /*
                 * ... markiere das aktuelle Kaestchen als besucht.
                 */
                besucht[aktuellX][aktuellY] = true;
                
                /*
                 * Setze bei allen Nachbarkstchen, die keine Wnde und noch
                 * nicht besucht sind, die Entfernung auf e + 1.
                 */
                if (gibLabyrinthinhalt (aktuellX, aktuellY + 1) != WAND
                            && !besucht[aktuellX][aktuellY + 1])
                {
                             entfernung[aktuellX][aktuellY + 1] = e + 1;
                }
                
                if (gibLabyrinthinhalt (aktuellX, aktuellY - 1) != WAND
                            && !besucht[aktuellX][aktuellY - 1])
                {
                             entfernung[aktuellX][aktuellY - 1] = e + 1;
                }
                
                if (gibLabyrinthinhalt (aktuellX + 1, aktuellY) != WAND
                            && !besucht[aktuellX + 1][aktuellY])
                {
                             entfernung[aktuellX + 1][aktuellY] = e + 1;
                }
                
                if (gibLabyrinthinhalt (aktuellX - 1, aktuellY) != WAND
                            && !besucht[aktuellX - 1][aktuellY])
                {
                             entfernung[aktuellX - 1][aktuellY] = e + 1;
                }  
            }

            /*
             * Solange die minimale Entfernung kleiner als unendlich ist,
             * wiederhole weiter.
             */
        }
        while (e < Double.POSITIVE_INFINITY);

        /*
         * Ende des Algorithmus von Dijkstra.
         * 
         * Ermittle die Anzahl der Leichen, Taeter, Zeugen und Feinde. 
         * Lsche dabei gleich diejenigen Personen, die vom Eingang aus 
         * nicht erreichbar sind.
         */
        leichen = 0;
        taeter = 0;
        zeugen = 0;
        feinde = 0;

        for (int x = 0; x < hoehe; x++)
        {
            for (int y = 0; y < breite; y++)
            {
                switch (labyrinth.gibFeld (x, y))
                {
                    case TAETER:
                    if (0 == taeter && entfernung[x][y] < Double.POSITIVE_INFINITY)
                    {
                        taeter++;
                    }
                    else
                    {
                        labyrinth.setzeFeld (x, y, LEERRAUM);
                    }
                    break;
                    
                    case LEICHE:
                    if (entfernung[x][y] < Double.POSITIVE_INFINITY)
                    {
                        leichen++;
                    }
                    else
                    {
                        labyrinth.setzeFeld (x, y, LEERRAUM);
                    }
                    break;
                    
                    case FEIND:
                    if (entfernung[x][y] < Double.POSITIVE_INFINITY)
                    {
                        feinde++;
                    }
                    else
                    {
                        labyrinth.setzeFeld (x, y, LEERRAUM);
                    }
                    break;
                }
                
                for (int k = 0; k < 9; k++)
                {
                    if (labyrinth.gibFeld (x, y) == ZEUGE[k])
                    {
                        if (entfernung[x][y] < Double.POSITIVE_INFINITY)
                    {
                        zeugen++;
                    }
                    else
                    {
                        labyrinth.setzeFeld (x, y, LEERRAUM);
                    }
                    }
                }
            }
        }
        
        if ( 0 == leichen && 0 == taeter)
        {
            System.out.println ( "Fehler." );
        }

        /*
         * Gehe zunaechst zum Eingang.
         */
        geheZumEingang ();
        
        /*
         * Durchlaufe das geladene Labyrinth und berprfe ob es dort eine abgespeicherte Position gibt.
         * Wenn diese Position vom Eingang aus erreichbar ist, dann nimm diese Position als Ausgangsposition
         * und blicke in die angegebene Richtung.
         */
        for (int x = 0; x < hoehe; x++)
        {
            for (int y = 0; y < breite; y++)
            {
                switch (labyrinth.gibFeld (x, y))
                {
                    case POSITIONNORD:
                    if (entfernung[x][y] < Double.POSITIVE_INFINITY)
                    {
                        positionX = x;
                        positionY = y;
                        blickrichtung = Himmelsrichtung.NORD;
                    }
                    labyrinth.setzeFeld (x, y, LEERRAUM);
                    break;
                    case POSITIONSUED:
                    if (entfernung[x][y] < Double.POSITIVE_INFINITY)
                    {
                        positionX = x;
                        positionY = y;
                        blickrichtung = Himmelsrichtung.SUED;
                    }
                    labyrinth.setzeFeld (x, y, LEERRAUM);
                    break;
                    case POSITIONWEST:
                    if (entfernung[x][y] < Double.POSITIVE_INFINITY)
                    {
                        positionX = x;
                        positionY = y;
                        blickrichtung = Himmelsrichtung.WEST;
                    }
                    labyrinth.setzeFeld (x, y, LEERRAUM);
                    break;
                    case POSITIONOST:
                    if (entfernung[x][y] < Double.POSITIVE_INFINITY)
                    {
                        positionX = x;
                        positionY = y;
                        blickrichtung = Himmelsrichtung.OST;
                    }
                    labyrinth.setzeFeld (x, y, LEERRAUM);
                    break;
                }
            }
        }

        /*
         * Erneuere die Labyrinthansicht.
         */
        erstelleAnsicht ();

        /*
         * Gib an die Beobachter weiter, wie viele Personen
         * im Labyrinth sind.
         */
        
        infotext = labyrinth.gibEinleitungstext ();
        naechstesBild = ziehe (tatortbild, tatortzahl);
        benachrichtigePersonenBeobachter ();
        
                               
        
    }
    
    /*
     * Alle Bilddateien, die im gesamten Spiel vorkommen.
     */
    private static String[] tatortbild  = {"IMG_1223.jpg"};
    private static String[] kumpanenbild = {"Anna, Kumpane.jpg", "chrissi,Kumpane.jpg", "Dina, Kumpane.jpg", 
        "Franzi,Kumpane.jpg", "Madlena,Kumpane.jpg", "Nicole,Kumpane.jpg", "Steffi, Kumpane.jpg"};
    private static String[] zeugenbild  = {"Co, lab1,2,3, 4.Zeuge.JPG", "Dani, lab1,2,3, 3.Zeuge.JPG",
        "Hannah, lab1,2,3, 1.Zeuge.JPG", "Inken,lab2, 2.Zeuge.JPG",
        "Kim,lab1,2,3, 5.Zeuge.JPG",
        "Laura, lab1,2,3, 6.Zeuge.JPG"};
    private static String[] opferbild  = {"feli, lab 1,2,3, oper 2.jpg", "lisa, lab1,2,3, opfer 1.jpg"};
    private static String[] taeterbild = {"IMG_1187b.JPG"};
    private static String[] schlussbild= {"IMG_1223.jpg"};
    
    /*
     * Zhlt, wie oft jedes Bild zufllig asugewhlt wurde, mit dem Ziel, dass jedes Bild gleich oft dran kommt.
     */
    private int[] tatortzahl   = new int[tatortbild.length];
    private int[] kumpanenzahl = new int[kumpanenbild.length];
    private int[] zeugenzahl   = new int[zeugenbild.length];
    private int[] opferzahl    = new int[opferbild.length];
    private int[] taeterzahl   = new int[taeterbild.length];
    private int[] schlusszahl  = new int[schlussbild.length];
    
    
    
    /**
     * Ziehe ohne Zuruecklegen aus dem Feld bildname.
     * Speichere dabei im Feld bildzahl, wie oft Du jeden bildnamen schon gezogen hast.
     * Wenn alle Bilder einmal dran waren, ziehe wieder aus dem gesamten Vorrat.
     */
    private String ziehe (String[] bildname, int[] bildzahl)
    {
        int laenge  = bildzahl.length;
        int minimum = bildzahl[0];
        int passend = 1;
        
        for (int i = 1; i < laenge; i++)
        {
           if (bildzahl[i] < minimum)
           {
               minimum = bildzahl[i];
               passend = 1;
           }
           else if (bildzahl[i] == minimum)
           {
               passend++;
           }
        }
        
        String ausgabe = "";
        
        int nummer = (int) Math.floor (passend * Math.random ());
        int j      = 0;
        
        for (int i = 0; i < laenge; i++)
        {
            if (bildzahl [i] == minimum)
            {
                if (j++ == nummer)
                {
                    ausgabe = bildname[i];
                    bildzahl[i] ++;
                }
            }
        }
        
        return ausgabe;
    }
    
    /**
     * Gibt das nchste Bild aus.
     */
    public String zeigeNaechstesBild ()
    {
        return naechstesBild;
    }
    
    /**
     * Gibt den nchsten Infotext aus.
     */
    public String gibNaechstenInfotext ()
    {
        return infotext;
    }

    
    /**
     * Gehe zum Eingang.
     * Drehe Dich dabei nach Moeglichkeit zu einem freien Gang hin.
     * Diese Methode wird am Ende des Ladens aufgerufen und wenn wir
     * einem Feind begegnet sind, der uns zurueck zum Eingang schickt.
     */
    public void geheZumEingang ()
    {
        /*
         * Suche den Eingang.
         */
        for (int x = 0; x < hoehe; x++)
        {
            for (int y = 0; y < breite; y++)
            {
                switch (labyrinth.gibFeld (x, y))
                {
                    case EINGANG:
                    positionX = x;
                    positionY = y;
                    break;
                }
            }
        }
        
        /*
         * Pruefe nacheinander im Norden, Westen und Sueden,
         * ob dort keine Wand ist. Wenn ja, blicke dorthin.
         * Wenn nein, blicke nach Osten.
         */
        if (WAND != gibLabyrinthinhalt (positionX, positionY + 1))
        {
            blickrichtung = Himmelsrichtung.NORD;
        }
        else if (WAND != gibLabyrinthinhalt (positionX - 1, positionY))
        {
            blickrichtung = Himmelsrichtung.WEST;
        }
        else if (WAND != gibLabyrinthinhalt (positionX, positionY - 1))
        {
            blickrichtung = Himmelsrichtung.SUED;
        }
        else
        {
            blickrichtung = Himmelsrichtung.OST;
        }
    }

    /**
     * Nenne die Seitenlaenge der Labyrinthansicht.
     */
    public int gibSeitenlaenge ()
    {
        return seitenlaenge;
    }

    /**
     * Gib die aktuelle Ansicht des Labyrinths.
     */
    public char[][] gibLabyrinthansicht ()
    {
        return ansicht;
    }

    /**
     * Gib die aktuelle Anzahl von Leichen.
     */
    public int gibLeichenzahl ()
    {
        return leichen;
    }

    /**
     * Gib die aktuelle Anzahl von Ttern.
     */
    public int gibTaeterzahl ()
    {
        return taeter;
    }

    /**
     * Gib die aktuelle Anzahl von Zeugen.
     */
    public int gibZeugenzahl ()
    {
        return zeugen;
    }

    /**
     * Gib die aktuelle Anzahl von Feinden.
     */
    public int gibFeindezahl ()
    {
        return feinde;
    }

    /**
     * Speichere einen Positionsbeobachter.
     * 
     * @param b Der Positionsbeobachter.
     */
    public void speichereAlsBeobachter (PositionsBeobachter b)
    {
        positionsBeobachter.add (b);
    }

    /**
     * Speichere einen Personenbeobachter.
     * 
     * @param b Der Personenbeobachter.
     */
    public void speichereAlsBeobachter (PersonenBeobachter b)
    {
        personenBeobachter.add (b);
    }

    /**
     * Loesche einen Positionsbeobachter.
     * 
     * @param b Der Positionsbeobachter.
     */
    public void loescheAlsBeobachter (PositionsBeobachter b)
    {
        positionsBeobachter.remove (b);
    }

    /**
     * Loesche einen Personenbeobachter.
     * 
     * @param b Der Personenbeobachter.
     */
    public void loescheAlsBeobachter (PersonenBeobachter b)
    {
        personenBeobachter.remove (b);
    }

    /**
     * Erstelle die duennwandige Ansicht.
     */
    private void erstelleAnsicht ()
    {
        /*
         * Fuelle die Ansicht zunaechst mit Leerraeumen.
         */
        for (int i = 0; i <= 4*seitenlaenge; i++)
        {
            for (int j = 0; j <= 4*seitenlaenge; j++)
            {
                ansicht[i][j] = LEERRAUM;;
            }
        }
        /*
         * Bestimme die Grenzen und die Schrittweiten der x-
         * und der y-Koordinaten.
         */
        int x1, x2, dx;
        int y1, y2, dy;

        switch (blickrichtung)
        {
            case NORD:
            x1 = positionX - seitenlaenge;
            x2 = positionX + seitenlaenge;
            dx = +1;
            y1 = positionY - seitenlaenge;
            y2 = positionY + seitenlaenge;
            dy = +1;
            break;

            case WEST:
            x1 = positionY - seitenlaenge;
            x2 = positionY + seitenlaenge;
            dx = +1;
            y1 = positionX + seitenlaenge;
            y2 = positionX - seitenlaenge;
            dy = -1;
            break;

            case SUED:
            x1 = positionX + seitenlaenge;
            x2 = positionX - seitenlaenge;
            dx = -1;
            y1 = positionY + seitenlaenge;
            y2 = positionY - seitenlaenge;
            dy = -1;
            break;

            default: // d. h. case OST:
            x1 = positionY + seitenlaenge;
            x2 = positionY - seitenlaenge;
            dx = -1;
            y1 = positionX - seitenlaenge;
            y2 = positionX + seitenlaenge;
            dy = +1;
            break;
        }

        /*
         * Verwende fuer die Ansicht die Koordinaten relativen
         * Koordinaten (u|w) zur Unterscheidung von dern
         * absoluten Koordinaten (x|y).
         */
        int u, v;

        /*
         * Trage alle Waende, Eingaenge und Schaetze korrekt in
         * die Ansicht ein.
         */
        u = -seitenlaenge;

        for (int x = x1; x != x2 + dx; x += dx)
        {
            v = -seitenlaenge;

            for (int y = y1; y != y2 + dy; y += dy)
            {
                char inhalt = WAND;

                switch (blickrichtung)
                {
                    case NORD: case SUED: inhalt = gibLabyrinthinhalt (x, y); break;
                    case OST:  case WEST: inhalt = gibLabyrinthinhalt (y, x); break;
                }

                switch (inhalt)
                {
                    /*
                     * Gruppiere vier Waende um ein Wandelement des
                     * dickwandigen Labyrinths herum.
                     */
                    case WAND:
                    setzeWand (index (u) + 1, index (v)    );
                    setzeWand (index (u)    , index (v) + 1);
                    setzeWand (index (u) - 1, index (v)    );
                    setzeWand (index (u)    , index (v) - 1);
                    break;

                    /*
                     * Uebertrage den Eingang.
                     */
                    case EINGANG:
                    ansicht[index (u)    ][index (v)    ] = EINGANG;
                    break;

                    /*
                     * Uebertrage die Feinde.
                     */
                    case FEIND:
                    ansicht[index (u)    ][index (v)    ] = FEIND;
                    break;

                    /*
                     * Uebertrage die Taeter.
                     */
                    case TAETER:
                    ansicht[index (u)    ][index (v)    ] = TAETER;
                    break;

                    /*
                     * Uebertrage die Leichen.
                     */
                    case LEICHE:
                    ansicht[index (u)    ][index (v)    ] = LEICHE;
                    break;
                    
                    /*
                     * Uebertrage die Marken in ihrer richtigen Orientierung.
                     */
                    case MARKENORD:
                    switch (blickrichtung)
                    {
                        case NORD: ansicht[index (u)][index (v)] = MARKENORD; break;
                        case WEST: ansicht[index (u)][index (v)] = MARKEOST;  break;
                        case SUED: ansicht[index (u)][index (v)] = MARKESUED; break;
                        case OST : ansicht[index (u)][index (v)] = MARKEWEST; break;
                    }
                    break;

                    case MARKEWEST:
                    switch (blickrichtung)
                    {
                        case NORD: ansicht[index (u)][index (v)] = MARKEWEST; break;
                        case WEST: ansicht[index (u)][index (v)] = MARKENORD; break;
                        case SUED: ansicht[index (u)][index (v)] = MARKEOST;  break;
                        case OST : ansicht[index (u)][index (v)] = MARKESUED; break;
                    }
                    break;

                    case MARKESUED:
                    switch (blickrichtung)
                    {
                        case NORD: ansicht[index (u)][index (v)] = MARKESUED; break;
                        case WEST: ansicht[index (u)][index (v)] = MARKEWEST;  break;
                        case SUED: ansicht[index (u)][index (v)] = MARKENORD; break;
                        case OST : ansicht[index (u)][index (v)] = MARKEOST;  break;
                    }
                    break;

                    case MARKEOST:
                    switch (blickrichtung)
                    {
                        case NORD: ansicht[index (u)][index (v)] = MARKEOST;  break;
                        case WEST: ansicht[index (u)][index (v)] = MARKESUED; break;
                        case SUED: ansicht[index (u)][index (v)] = MARKEWEST; break;
                        case OST : ansicht[index (u)][index (v)] = MARKENORD; break;
                    }
                    break;    
                }
                
                /*
                 * Uebertrage die Zeugen.
                 */
                for (int k = 0; k < 9; k++)
                {
                    if (inhalt == ZEUGE [k])
                    {
                        ansicht[index (u)    ][index (v)    ] = ZEUGE [k];
                    }
                }
                
                v++;
            }
            u++;
        }

        benachrichtigePositionsBeobachter ();
    }

    /**
     * Pruefe, ob das Spielende erreicht ist.
     */
    private void pruefeSpielende ()
    {
        /*
         * Wenn ich alle Leichen und Taeter eingesammelt habe und zurueck
         * auf dem Eingang bin, dann vermindere ich die
         * Taeter-&Leichenzahl nochmals um 1, um anzuzeigen, dass das
         * Spiel zu Ende ist.
         */
        if (EINGANG == gibLabyrinthinhalt (positionX, positionY) && 0 == taeter && 0 ==  leichen)
        {
            taeter --;
            leichen --;
            naechstesBild = ziehe (schlussbild, schlusszahl);
            infotext = "Du hast gewonnen! Herzlichen Glckwunsch!!";
            benachrichtigePersonenBeobachter ();;
        } 
    }

    /**
     * Nenne den Inhalt des Labyrinths an der Position
     * (x|y). Wenn die Position unmittelbar jenseits des
     * Labyrinthrandes ist, tun wir so, als waere dort
     * eine Wand. Wenn die Position noch weiter ausserhalb
     * des Labyrints liegt, tun wir so, als waere dort ein
     * Leerraum.
     */
    private char gibLabyrinthinhalt (int x, int y)
    {
        if (0 <= x && x < hoehe)
        {
            if (0 <= y && y < breite)
            {
                return labyrinth.gibFeld (x, y);
            }
            else if (-1 == y || breite == y)
            {
                return WAND;
            }
            else
            {
                return LEERRAUM;
            }
        }
        else if (-1 == x || hoehe == x)
        {
            if (0 <= y && y < breite)
            {
                return WAND;
            }
            else
            {
                return LEERRAUM;
            }
        }
        else
        {
            return LEERRAUM;
        }
    }

    /**
     * Rechne eine Koordinate des Labyrinths in einen
     * Feldindex der Ansicht um.
     */
    private int index (int k)
    {
        return 2*seitenlaenge - 2*k;
    }

    /**
     * Setze an der angegebenen Stelle eine Wand in die Ansicht.
     */
    private void setzeWand (int i, int j)
    {
        /*
         * Pruefe, ob die Stelle innerhalb der Ansicht liegt.
         */
        if (0 <= Math.min (i, j) && Math.max (i, j) <= 4*seitenlaenge)
        {
            if (WAND != ansicht[i][j])
            {
                /*
                 * Wenn an der Stelle noch keine Wand ist, setze eine.
                 */
                ansicht[i][j] = WAND;
            }
            else
            {
                /*
                 * Wenn an der Stelle bereits eine Wand ist, loesche
                 * sie (denn dann stossen dort zwei Wandelemente
                 * aneinander).
                 */
                ansicht[i][j] = LEERRAUM;
            }
        }   
    }

    /**
     * Benachrichtige Positionsbeobachter, dass sich an der Position
     * im Labyrinth etwas geaendert hat.
     */
    private void benachrichtigePositionsBeobachter ()
    {
        for (PositionsBeobachter b: positionsBeobachter)
        {
            b.aktualisiereLabyrinthansicht ();
        }
    }

    /**
     * Benachrichtige Personenbeobachter, dass sich an der Anzahl der
     * Personen im Labyrinth etwas geaendert hat.
     */
    private void benachrichtigePersonenBeobachter ()
    {
        for (PersonenBeobachter b: personenBeobachter)
        {
            b.aktualisierePersonenzahl ();
            b.zeigeNaechstesBild ();
        }
    }
}