import java.awt.Color;
import java.awt.Point;
import java.lang.IllegalArgumentException;

/**
 * Eine hintere (horizontale) Wandkachel im Labyrinth.
 * 
 * @author Bernd Gramlich
 * @version 2013-01-04
 */
public class HintereWandkachel extends Kachel
{
    /**
     * Konstruktor.
     * Erzeuge eine hintere Wandkachel an der angegebenen
     * Position mit der angegebenen Faerbung. Die Ausgabeleinwand
     * hat dabei die angegebene Hoehe und Breite und die
     * angegebenen Fluchtpunktkoordinaten.
     * 
     * Eine hintere Wandkachel hat immer geradzahlige x-Koordinate
     * und ungeradzahlige y-Koordinate. Die y-Koordinate ist nicht
     * negativ. Wenn die Argumente x und y diese Bedingung nicht
     * erfuellen, melde einen Fehler.
     */
    public HintereWandkachel (int x, int y, Color dieseFaerbung, int breite, int hoehe, int xf, int yf, double mu, double weite)
        throws IllegalArgumentException
    {
        super (x, y, dieseFaerbung, breite, hoehe, xf, yf, mu, weite);
        
        /*
         * Wenn die x-Koordinate ungerade ist, melde einen Fehler.
         */
        if (x % 2 == 1)
        {
            throw new IllegalArgumentException ("Hintere Wandkacheln haben geradzahlige x-Koordinate!");
        }
        
        /*
         * Wenn die y-Koordinate gerade oder negativ ist, melde einen Fehler.
         */
        if (y % 2 == 0 || y < 0)
        {
            throw new IllegalArgumentException ("Hintere Wandkacheln haben ungerade, nichtnegative y-Koordinate!");
        }
        
        abstand = Math.abs (x) + y;
        
        euklidischerAbstand = Math.sqrt (Math.pow (x, 2) + Math.pow (y, 2));
        
        if (0 == x)
        {
            dP = Math.sqrt (Math.pow (0, 2) + Math.pow (y, 2));
            dQ = Math.sqrt (Math.pow (1, 2) + Math.pow (y, 2));
            zyklisch = true;
        }
        else
        {
            dP = Math.sqrt (Math.pow (x - 1, 2) + Math.pow (y, 2));
            dQ = Math.sqrt (Math.pow (x + 1, 2) + Math.pow (y, 2));
            zyklisch = false;
        }
        
        aendereLeinwanddaten (dieseFaerbung, breite, hoehe, xf, yf, mu, weite);
    }
    
    /**
     * Aendere die Leinwanddaten.
     * Die hintere Wandkachel muss entsprechend den neuen Leinwanddaten
     * geformt und gefaerbt werden.
     */
    public void aendereLeinwanddaten (Color dieseFaerbung, int breite, int hoehe, int xf, int yf, double mu, double weite)
    {
        double lambda = mu * (2 - mu);
        
        bild.loescheAlleEcken ();
        
        int dx = positionX / 2;
        int dy = (positionY - 1) / 2;
        
        double faktor = 1 - (1 - mu) * Math.pow (1 - lambda, dy);
            
        int xl = (int) Math.round (faktor * xf);
        int xr = (int) (breite - Math.round (faktor * (breite - xf)));
        int yu = (int) (hoehe - Math.round (faktor * (hoehe - yf)));
        int yo = (int) Math.round (faktor * yf);
            
        bild.ergaenzeEcke (dx * (xr - xl) + xl, yu);
        bild.ergaenzeEcke (dx * (xr - xl) + xr, yu);
        bild.ergaenzeEcke (dx * (xr - xl) + xr, yo);
        bild.ergaenzeEcke (dx * (xr - xl) + xl, yo);
        
        if (0 == dx)
        {
            P = new Point ((xr + xl)/2, (yu + yo) / 2);
        }
        else
        {
            P = new Point (dx * (xr - xl) + xl, (yu + yo) / 2);
        }
        
        Q = new Point (dx * (xr - xl) + xr, (yu + yo) / 2);
        
        hellsichtweite = weite;
        faerbung = dieseFaerbung;
        bild.setzeFarbe (erzeugeBildfaerbung ());
    }
    
    /**
     * Pruefe, ob die Kachel eine Wandkachel ist.
     */
    public boolean istWandkachel ()
    {
        return true;
    }
}