/**
 * Das Labyrinth, bestehend aus 
 * - der Karte,
 * - den Bemerkungen der einzelnen Zeugen.
 * 
 * @author Felicia W.
 * @version 2013-06-07
 */
public class Labyrinth
{
    /*
     * Die Karte des Labyrinths speichern wir als rechteckiges Feld von Textzeichen.
     */    
    private char [][] karte;
    
    /*
     * Die Bemerkungen der Zeugen speichern wir in einem Feld von Zeichenketten.
     */
    private String[] zeugenbemerkung;
    
    /*
     * So sieht ein Eingang aus.
     */
    private final char EINGANG      = '@';
    
    /*
     * Das Minimallabyrinth, das nur aus einem Eingang besteht.
     */
    private final char[][] minimalLabyrinth = {{EINGANG}};
    
    /*
     * Der Einleitungstext als Zeichenkette.
     */
    private String einleitungstext;


    /**
     * Konstruktor
     * Erzeuge ein feld fr Bemekungen der Zeugen 1 bis 9.
     * Verwende als Karte zunchst das minimal mgliche Labyrinth.
     */
    public Labyrinth()
    {
        zeugenbemerkung = new String[10];
        
        karte = minimalLabyrinth;
    }
    
    /**
     * Gib dem Labyrinth eine neue Karte.
     */
    public void setzeKarte (char[][] neueKarte)
    {
        karte = neueKarte;
    }
    
    /**
     * Nenne die Breite des Labyrinths.
     * (Wir gehen dabei davon aus, dass das Labyrinth reckteckig ist.)
     */
    public int gibBreite ()
    {
        return karte[0].length;
    }
    
    /**
     * Nenne die Hhe des Labyrinths.
     */
    public int gibHoehe ()
    {
        return karte.length;
    }
    
    /**
     * Nenne den Inhalt des Kartenfeldes mit den Koordinaten (x|y).
     * Eine 0 als Rckgabewert zeigt an, dass die Joordinaten ungltig waren.
     */
    public char gibFeld (int x, int y)
    {
        if (x >= 0 && x <= gibHoehe () && y >= 0 && y <= gibBreite ())
        {
            return karte[x][y];
        }
        else
        {
            return 0;
        }
    }
    
    /**
     * Lege die Bemerkung des Zeugen Nummer i fest.
     */
    public void setzeZeugenaussage (int i, String aussage)
    {
        if (i >= 1 && i <= 9)
        {
            zeugenbemerkung [i] = aussage;
        }
    }
    
    /**
     * Nenne die Bemerkung des Zeugen Nummer i.
     */
    public String gibZeugenaussage (int i)
    {
         if (i >= 1 && i <= 9)
        {
            return zeugenbemerkung [i];
        }
        else
        {
            return "";
        }
    }
    
    /**
     * ndere den Inhalt des Kartenfeldes mit den Koordinaten (x|y).
     */
    public void setzeFeld (int x, int y, char zeichen)
    {
        if (x >= 0 && x <= gibHoehe () && y >= 0 && y <= gibBreite ())
        {
           karte[x][y] = zeichen;
        }
    }
    
    /**
     * Setze den Inhalt des Einleitungstextes.
     */
    public void setzeEinleitungstext (String aussage)
    {
        einleitungstext = aussage;
    }
    
    /**
     * Gib den Inhalt des Einleitungstextes aus.
     */
    public String gibEinleitungstext ()
    {
        return einleitungstext;
    }
}
