import java.awt.Color;
import java.awt.Point;
import java.lang.IllegalArgumentException;

/**
 * Eine rechte Wandkachel im Labyrinth.
 * 
 * @author Bernd Gramlich
 * @version 2013-01-04
 */
public class RechteWandkachel extends Kachel
{
    /**
     * Konstruktor.
     * Erzeuge eine rechte Wandkachel an der angegebenen Position
     * mit der angegebenen Faerbung. Die Ausgabeleinwand hat
     * dabei die angegebene Hoehe und Breite und die angegebenen
     * Fluchtpunktkoordinaten.
     * 
     * Eine rechte Wandkachel hat immer ungeradzahlige x-Koordinate
     * und geradzahlige y-Koordinate. Die x-Koordinate ist positiv,
     * die y-Koordinate ist nicht negativ. Wenn die Argumente x und
     * y diese Bedingung nicht erfuellen, melde einen Fehler.
     */
    public RechteWandkachel (int x, int y, Color dieseFaerbung, int breite, int hoehe, int xf, int yf, double mu, double weite)
        throws IllegalArgumentException
    {
        super (x, y, dieseFaerbung, breite, hoehe, xf, yf, mu, weite);
        
        /*
         * Wenn die x-Koordinate gerade oder negativ ist, melde einen Fehler.
         */
        if (x % 2 == 0 || x < 0)
        {
            throw new IllegalArgumentException ("Rechte Wandkacheln haben ungeradzahlige, positive x-Koordinate!");
        }
        
        /*
         * Wenn die y-Koordinate ungerade oder negativ ist, melde einen Fehler.
         */
        if (y % 2 == 1 || y < 0)
        {
            throw new IllegalArgumentException ("Rechte Wandkacheln haben gerade, nichtnegative y-Koordinate!");
        }
        
        abstand = Math.abs (x) + y;
        
        euklidischerAbstand = Math.sqrt (Math.pow (x, 2) + Math.pow (y, 2));
        
        dP = Math.sqrt (Math.pow (x, 2) + Math.pow (y - 1, 2));
        dQ = Math.sqrt (Math.pow (x, 2) + Math.pow (y + 1, 2));
        zyklisch = false;
        
        aendereLeinwanddaten (dieseFaerbung, breite, hoehe, xf, yf, mu, weite);
    }
    
    /**
     * Aendere die Leinwanddaten.
     * Die rechte Wandkachel muss entsprechend den neuen Leinwanddaten
     * geformt und gefaerbt werden.
     */
    public void aendereLeinwanddaten (Color dieseFaerbung, int breite, int hoehe, int xf, int yf, double mu, double weite)
    {
        double lambda = mu * (2 - mu);
        
        bild.loescheAlleEcken ();
        
        int dx = (positionX + 1) / 2;
        int dy = positionY / 2;
        
        double faktorv = 1 - (1 - mu) * Math.pow (1 - lambda, dy - 1);
        double faktorh = 1 - (1 - mu) * Math.pow (1 - lambda, dy);
        
        if (0 == dy)
        {
            faktorv = 0;
        }
            
        int xl = (int) Math.round (dx * breite - faktorh * (dx * breite - xf));
        int xr = (int) Math.round (dx * breite - faktorv * (dx * breite - xf));
        int yul  = (int) Math.round (hoehe - faktorh * (hoehe - yf));
        int yur  = (int) Math.round (hoehe - faktorv * (hoehe - yf));
        int yol  = (int) Math.round (faktorh * yf);
        int yor  = (int) Math.round (faktorv * yf);
            
        bild.ergaenzeEcke (xl, yul);
        bild.ergaenzeEcke (xr, yur);
        bild.ergaenzeEcke (xr, yor);
        bild.ergaenzeEcke (xl, yol);
        
        P = new Point (xr, (yor + yur)/2);
        Q = new Point (xl, (yol + yul)/2);
        
        hellsichtweite = weite;
        faerbung = dieseFaerbung;
        bild.setzeFarbe (erzeugeBildfaerbung ());
    }
    
    /**
     * Pruefe, ob die Kachel eine Wandkachel ist.
     */
    public boolean istWandkachel ()
    {
        return true;
    }
}